--------------------------------------------------------------------------------
-- From Deriving to Regular
--------------------------------------------------------------------------------

{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE PolyKinds              #-}
{-# LANGUAGE TypeFamilies           #-}
{-# LANGUAGE TypeOperators          #-}
{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE UndecidableInstances   #-}
{-# LANGUAGE MultiParamTypeClasses  #-}

module Conversion.D2R where


import qualified Library.Deriving as D
import qualified Library.Regular  as R

--------------------------------------------------------------------------------
-- Conversion
--------------------------------------------------------------------------------

-- Type conversion
type family D2R (u :: D.Un *) :: R.Un *

type instance D2R D.UU           = R.U
type instance D2R ((D.:+:)  a b) = R.SUM  (D2R a) (D2R b)
type instance D2R ((D.:**:) a b) = R.PROD (D2R a) (D2R b)
type instance D2R (D.MM c a)     = R.C c (D2R a)

-- The interesting case is D.KK, and here we see how a mere tag (generated by
-- the compiler when deriving the representation for a datatype) can prove
-- useful:
type instance D2R (D.KK (D.R D.S) r)  = R.I
type instance D2R (D.KK (D.R D.O) t)  = R.K t
type instance D2R (D.KK D.P a)        = R.K a
type instance D2R (D.KK D.U a)        = R.K a

-- The remaining D representation types (PAR, REC, and :.:) are not considered
-- because we will only convert Generic representations, not Generic1.


-- Value conversion
class Convert (u :: D.Un *) r where
  d2r :: D.In u p -> R.In (D2R u) r
  r2d :: R.In (D2R u) r -> D.In u p

instance Convert D.VV r where
  d2r = error "no values of type VV"
  r2d = error "no values of type VV"

instance Convert D.UU r where
  d2r _ = R.In_U
  r2d _ = D.U1

instance (Convert a r) => Convert (D.MM c a) r where
  d2r (D.M1 a)   = R.In_C (d2r a)
  r2d (R.In_C a) = D.M1 (r2d a)

instance Convert (D.KK (D.R D.S) r) r where
  d2r (D.K1 a)   = R.In_I a
  r2d (R.In_I a) = D.K1 a

instance Convert (D.KK (D.R D.O) t) r where
  d2r (D.K1 a)   = R.In_K a
  r2d (R.In_K a) = D.K1 a

instance Convert (D.KK D.P a) r where
  d2r (D.K1 a)   = R.In_K a
  r2d (R.In_K a) = D.K1 a

instance Convert (D.KK D.U a) r where
  d2r (D.K1 a)   = R.In_K a
  r2d (R.In_K a) = D.K1 a

instance (Convert a r, Convert b r) => Convert ((D.:+:) a b) r where
  d2r (D.L1 a)   = R.In_L (d2r a)
  d2r (D.R1 a)   = R.In_R (d2r a)
  r2d (R.In_L a) = D.L1 (r2d a)
  r2d (R.In_R a) = D.R1 (r2d a)

instance (Convert a r, Convert b r) => Convert ((D.:**:) a b) r where
  d2r ((D.:*:) a b) = R.In_P  (d2r a) (d2r b)
  r2d (R.In_P  a b) = (D.:*:) (r2d a) (r2d b)

--------------------------------------------------------------------------------
-- Embedding conversion
--------------------------------------------------------------------------------

-- All Deriving types are representable in Regular
instance (D.Generic a, Convert (D.Rep a) a) => R.Regular a where
  type PF a = D2R (D.Rep a)

  from x = d2r (D.from x)
  to   x = D.to (r2d x)

-- Note that this is the most general and only instance that can be given for
-- this class, but that's ok, because that's exactly what we want: never to have
-- to write instances for R.Regular, simply use the stuff from D.Generic.
